# dxf2ruby.rb - (C) 2011 jim.foltz@gmail.com
# $Id: dxf2ruby.rb rev:21 node:230a140f1deb branch:default 2011-06-21 $

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# == Purpose ==
#
# To parse .dxf files into something useful by Ruby using only built-in Ruby
# objects (Hash, Array, Float, Fixnum, String). Meant to be a building-block
# for more advanced DXF processing.


# == Quick Start ==
#
#   require 'dxf2ruby'
#   dxf = Dxf2Ruby.parse("file.dxf")
#   acad_version =  dxf['HEADER']['$ACADVER']
#   dxf['ENTITIES'].each do |entity|
#      draw(entity)
#   end
#
#   def draw(entity)
#       case entity[0]
#         when "POINT"
#           draw_point(entity)
#        when "LINE"
#           draw_line(entity)
#      end
#   end
#
# Dxf2RUby may also be used on a command line:
#
#       $ dxf2ruby file.dxf

# == About ==
#
# dxf2ruby returns a Ruby data structure from a .dxf file. Dxf2Ruby does not
# attempt to interpret the codes, but rather assembles the codes into Acad
# objects in the form of Ruby Hashes.
#
# The top-level data structure is a Hash of containers representing SECTIONS of
# the .dxf file.  Currently supported sections are the HEADER, BLOCKS, and
# ENTITIES sections.
#
#   {"HEADER"=>{...}, "BLOCKS"=>[...], "ENTITIES"=>[...]}


# == The HEADER Section ==
#
#    The HEADER section is a Ruby Hash object. Reading the HEADER variables can
#    be done as follows:
#
#       header = dxf['HEADER']
#       acad_version = header['$ACADVER'][1]
#
#    Non-existant variable return nil.


# == The BLOCKS Section ==
#
# The BLOCKS section is a Ruby Array.


# == The ENTITIES Section ==
#
# Acad entities are returned as Ruby Hashes.  A typical entity may look similar
# to the following Hash, which represents an Acad LINE entity:
#
#   {0=>"LINE", 5=>"25", 100=>["AcDbEntity", "AcDbLine"], 8=>"0",
#   6=>"CONTINUOUS", 62=>7, 10=>-4.672884, 20=>-0.816414, 30=>0.0,
#   11=>-4.672884, 21=>-0.27178, 31=>0.0}


# About DXF (Nots from AutoCAD)
#
# DXF _Objects_ have no graphical representation. (aka nongraphical objects)
# DXF _Entities_ are graphical objects.

# Accommodating DXF files from future releases of AutoCAD will be easier
# if you write your DXF processing program in a table-driven way, ignore
# undefined group codes, and make no assumptions about the order of group codes
# in an entity. With each new AutoCAD release, new group codes will be added to
# entities to accommodate additional features.

module Dxf2Ruby

    module_function

    def parse(filename)
        @debug = $JFDEBUG
        fp = File.open(filename)

        dxf       = {'HEADER' => {}, 'BLOCKS' => [], 'ENTITIES' => [] }
        last_code = nil
        last_ent  = nil

        #
        # main loop
        #

        while true
            c, v = read_codes(fp)
            break if v == "EOF"
            if v == "SECTION"
                c, v = read_codes(fp)

                if v == "HEADER"
                    hdr = dxf['HEADER']
                    while true
                        c, v = read_codes(fp)
                        break if v == "ENDSEC"
                        if c == 9
                            key = v
                            hdr[key] = {}
                        else
                            add_att(hdr[key], c, v)
                        end
                    end # while
                end # if HEADER

                if v == "BLOCKS"
                    blks = dxf[v]
                    parse_entities(blks, fp)
                end # if BLOCKS

                if v == "ENTITIES"
                    ents = dxf[v]
                    parse_entities(ents, fp)
                end #  ENTITIES section

            end # if in SECTION

        end # main loop

        return dxf
    end

    def parse_entities(section, fp)
        last_ent = nil
        last_code = nil
        while true
            c, v = read_codes(fp)
            break if v == "ENDSEC"
            next if c == 999
            if last_ent == "LWPOLYLINE"
                if last_code == 20 and c != 42
                    add_att(section[-1], 42, 0.0)
                end
            end
            if c == 0
                last_ent = v
                section << {c => v}
            else
                add_att(section[-1], c, v)
            end
            last_code = c
        end # while
    end # def parse_entities

    def read_codes(fp)
        c = fp.gets.to_i
        v = fp.gets.strip
        v.upcase! if c == 0
        case c
        when 10..59, 140..147, 210..239, 1010..1059
            v = v.to_f
        when 60..79, 90..99, 170..175,280..289, 370..379, 380..389,500..409, 1060..1079
            v = v.to_i
        end
        return( [c, v] )
    end

    def add_att(ent, code, value)
        if ent.nil? and @debug
            p caller
            p code
            p value
        end
        if ent[code].nil?
            ent[code] = value
        elsif ent[code].class == Array
            ent[code] << value
        else
            t = ent[code]
            ent[code] = []
            ent[code] << t
            ent[code] << value
        end
    end


end # class Dxf2Ruby


if $0 == __FILE__
    require 'pp'
    t1 = Time.now
    dxf = Dxf2Ruby.parse(ARGV.shift)
    puts "Finsihed in #{Time.now - t1}"
    pp dxf
end
